
;;  Paradigmata Programovani 2B, PRIKLADY
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;; 1. CAST: LINE VYHODNOCENI
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; LINE VYHODNOCOVANI: prisliby
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; zakladni myslenka: mame k dispozici formu `delay', ktera vraci prislib
;;                    budouciho vyhodnoceni daneho vyrazu
;;                    dale mame proceduru `force', ktera pro dany prislib
;;                    aktivuje vypocet a vrati hodnotu vzniklou vyhodnocenim
;;                    prislibeneho vyrazu
;;                    tento princip se nazyva ``call by need''

;; jednoduch ukazka
(delay (+ 1 2))        ; #<struct:promise>
(define p (delay (map - '(1 2 3 4))))
p
(promise? p)
(force p)        ; (-1 -2 -3 -4)

;; Poznamka: delay nemuze byt procedura

;; u prislibenych vyrazu dochazi k vypoctu az pri jejich aktivaci pomoci force
;; to plati i o chybach

(define p (delay blah))        ; probehne bez problemu
p                        ; #<struct:promise>
(force p)                ; CHYBA (projevi se az zde!)

;; pomoci prislibu je mozne ,,odlozit slozity vypocet na pozdeji''
;; az je ,,skutecne potreba jej provest''

;; casove narocny vypocet:
(define fib
  (lambda (n)
    (if (<= n 2)
        1
        (+ (fib (- n 1))
           (fib (- n 2))))))

(define p (delay (fib 30)))        ; probehne okamzite
(force p)                        ; az zde se skutecne pocita (prodleva)

;; pri pouziti lineho vyhodnoceni vyvstavaji otazky spojene s vedlejsim efektem

(define p (let ((i 0))
            (delay (begin
                     (set! i (+ i 1))
                     i))))
(force p)        ; 1
(force p)        ; ???

;; R5RS: The effect of <expression> returning multiple values is unspecified.
;; takze pri druhem volani force na tyz prislib specifikace Scheme nijak
;; neupracuje, jaka je vysledna hodnota. Mame dve zakladni moznosti.

;; 1) ``force'' bude porzne vyhodnocovat prislibeny vyraz
;;      klady: chova se dobre k vedlejsim efektum, zapor: opakovane vyhodnoceni
;; 2) ``force'' prislibeny vyraz vyhodnoti jen jednou, vyslednou hodnotu
;;    si uchovava a pri dalsich spustenich vraci jen tuto hodnotu
;;      klady: pri dalsich vyhodnocenich mame vysledek ,,okamzite''
;;      zapor: pro procedury s vedlejsim efektem je to neprirozene

;; V pripade 1) budeme mit:
(force p)        ; 1
(force p)        ; 2
(force p)        ; 3 ...

;; V pripade 2) budeme mit:
(force p)        ; 1
(force p)        ; 1
(force p)        ; 1 ...

;; Poznamka. Vetsina interpretu voli cestu 2)

;;; IMPLEMENTACE PRISLIBU
;;;;;;;;;;;;;;;;;;;;;;;;;

;; Prisliby lze plne implementovat pomoci
;; procedur vyssich radu, maker a vedlejsich efektu

;; zakladni myslenka: pri vytvareni procedur nedochazi k vyhodnocovani jejich
;; tela, k tomu dochazi az pri jejich aplikaci .. nabizi se tedy vytvorit
;; prisliby pomoci procedur

(lambda () (+ 1 2))                ; #<procedure>
(define p (lambda () (+ 1 2)))        ; nas prislib
(p)                                ; aktivace

;; implementace 1): freeze & thaw
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; specialni forma ``freeze''
(define-macro freeze
  (lambda exprs
    `(lambda ()
       (begin ,@exprs))))

;; procedura ``thaw''
(define thaw
  (lambda (promise)
    (promise)))

; priklad pouziti:
(define p
  (let ((x 10))
    (freeze (display "Hodnota: ")
            (display x)
            (newline)
            (set! x (+ x 1))
            (list x (* x x)))))

(thaw p)        ; (11 121)
(thaw p)        ; (12 144)
(thaw p)        ; (13 169) ...

;; implementace 2): delay & force
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; specialni forma ``delay''
(define-macro delay
  (lambda exprs
    `(let ((result (lambda ()
                     (begin ,@exprs)))
           (evaluated? #f))
       (lambda ()
         (begin
           (if (not evaluated?)
               (begin
                 (set! evaluated? #t)
                 (set! result (result))))
           result)))))

;; procedura ``force'' (totez to ``thaw'')
(define force thaw)

; priklad pouziti:
(define p
  (let ((x 10))
    (delay (display "Hodnota: ")
           (display x)
           (newline)
           (set! x (+ x 1))
           (list x (* x x)))))

(force p)        ; (11 121)
(force p)        ; (11 121)
(force p)        ; (11 121) ... + nic se nevypisuje na obrazovku

;; 2. CAST: PROUDY (STREAMS)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; !!!! JINAK SE URCUJI SLOZITOSTI!

;; potencialne nekonecne datove struktury
;; nekonecne a implicitni proudy
;; manipulace s proudy

;; konstruktor proudu (je specialni forma)
;; doted byly vsechny datove konstruktory procedury
(define-macro cons-stream
  (lambda (a b)
    `(cons ,a (delay ,b))))

;; selektor stream-car (vrat prvni prvek proudu)
(define stream-car car)

;; selektor stream-cdr (vrat proud bez prvniho prvku)
(define stream-cdr
  (lambda (stream)
    (force (cdr stream))))

;; predikat: predikat stream?
(define stream?
  (lambda (elem)
    (or (null? elem)
        (and (pair? elem)
             (or (promise? (cdr elem))
                 (null? (cdr elem)))))))

(define stream?
  (lambda (elem)
    (or (null? elem)
	(and (pair? elem)
	     (and (promise? (cdr elem))
		  (stream? (force (cdr elem))))))))

;; predikat: je stream prazdny?
(define stream-null? null?)

;; zobraz stream, nanejvys vsak n prvnich prvku
(define display-stream
  (lambda (stream . n)
    (let iter ((f #f)
               (stream stream)
               (n (if (null? n)
                      #f
                      (car n))))
      (if (and (or (not n) (> n 0))
               (not (stream-null? stream)))
          (begin
            (if f
                (display " ")
                (display "#<stream ("))
            (display (stream-car stream))
            (iter #t
                  (stream-cdr stream)
                  (if (not n)
                      n
                      (- n 1))))
          (begin
            (display ")>")
            (newline))))))

(define stream-caar (lambda (x) (stream-car (stream-car x))))
(define stream-cadr (lambda (x) (stream-car (stream-cdr x))))
(define stream-cdar (lambda (x) (stream-cdr (stream-car x))))
(define stream-cddr (lambda (x) (stream-cdr (stream-cdr x))))
(define stream-caaar (lambda (x) (stream-car (stream-caar x))))
(define stream-caadr (lambda (x) (stream-car (stream-cadr x))))
(define stream-cadar (lambda (x) (stream-car (stream-cdar x))))
(define stream-caddr (lambda (x) (stream-car (stream-cddr x))))
(define stream-cdaar (lambda (x) (stream-cdr (stream-caar x))))
(define stream-cdadr (lambda (x) (stream-cdr (stream-cadr x))))
(define stream-cddar (lambda (x) (stream-cdr (stream-cdar x))))
(define stream-cdddr (lambda (x) (stream-cdr (stream-cddr x))))
(define stream-caaaar (lambda (x) (stream-caar (stream-caar x))))
(define stream-caaadr (lambda (x) (stream-caar (stream-cadr x))))
(define stream-caadar (lambda (x) (stream-caar (stream-cdar x))))
(define stream-caaddr (lambda (x) (stream-caar (stream-cddr x))))
(define stream-cadaar (lambda (x) (stream-cadr (stream-caar x))))
(define stream-cadadr (lambda (x) (stream-cadr (stream-cadr x))))
(define stream-caddar (lambda (x) (stream-cadr (stream-cdar x))))
(define stream-cadddr (lambda (x) (stream-cadr (stream-cddr x))))
(define stream-cdaaar (lambda (x) (stream-cdar (stream-caar x))))
(define stream-cdaadr (lambda (x) (stream-cdar (stream-cadr x))))
(define stream-cdadar (lambda (x) (stream-cdar (stream-cdar x))))
(define stream-cdaddr (lambda (x) (stream-cdar (stream-cddr x))))
(define stream-cddaar (lambda (x) (stream-cddr (stream-caar x))))
(define stream-cddadr (lambda (x) (stream-cddr (stream-cadr x))))
(define stream-cdddar (lambda (x) (stream-cddr (stream-cdar x))))
(define stream-cddddr (lambda (x) (stream-cddr (stream-cddr x))))

;; delka streamu
(define stream-length
  (lambda (stream)
    (if (stream-null? stream)
        0
        (+ 1 (stream-length (stream-cdr stream))))))

;; mapovani pres streamy (mapovani pres jeden stream)
(define stream-map2
  (lambda (f stream)
    (if (stream-null? stream)
        '()
        (cons-stream 
         (f (stream-car stream))
         (stream-map f (stream-cdr stream))))))

;; mapovani pres streamy (obecna verze)
(define stream-map
  (lambda (f . streams)
    (if (stream-null? (car streams))
        '()
        (cons-stream
         (apply f (map stream-car streams))
         (apply stream-map f
                (map stream-cdr streams))))))

;; foldr pro streamy
;; tady je chytak: tohle nefunguje
(define stream-foldr
  (lambda (f nil . streams)
    (if (stream-null? (car streams))
        nil
        (apply 
	 f
	 `(,@(map stream-car streams)
	   ,(apply stream-foldr f nil
		   (map stream-cdr streams)))))))

;; zdanlive funkcni
(stream-foldr (lambda (x y)
		(cons-stream (- x) (force y)))
	      '()
	      (stream 1 2 3 'blah 4))	; error

;; reseni: vytvorime foldr, ve kterem bude druhy argument procedury f prislib
;; pri pouziti foldr si to budeme muset jako uzivatele hlidat
(define stream-foldr
  (lambda (f nil . streams)
    (if (stream-null? (car streams))
        nil
        (apply
         f
         `(,@(map stream-car streams)
             ,(delay
                 (apply stream-foldr f nil
                        (map stream-cdr streams))))))))

; priklad pouziti
(stream-foldr (lambda (x y)
                (cons-stream (- x) (force y)))
              '()
              (build-stream (lambda (i) i)))

;; ODVOZENE POMOCNE PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konvertuj seznam na stream
(define list->stream
  (lambda (list)
    (foldr (lambda (x y)
             (cons-stream x y))
           '() list)))

;; vytvor konecny stream danych argumentu
(define stream
  (lambda args
    (list->stream args)))

(define s (stream 1 2 3 4))
s                                      ; (1 . #<promise>)
(display-stream s)                     ; #<stream (1 2 3 4)>
(display-stream s 2)                   ; #<stream (1 2)>
(stream-length s)                      ; 4
(display-stream (stream-map - s) 2)    ; #<stream (-1 -2)>
(display-stream (stream-map + s s s))  ; #<stream (3 6 9 12)>

(define s (stream 1 2 3 4 5 'blah 6 7))
(define r (stream-map - s))
r                     ; (-1 . #<promise>)
(display-stream r 4)  ; (-1 -2 -3 -4)
(display-stream r 6)  ; (-1 -2 -3 -4 -5 error

(define fs (stream-map fib (stream 1 ... 30 31 32 ... 50)))
fs  ; (1 . #<promise>)

;; konverze streamu na seznam
(define stream->list
  (lambda (stream)
    (stream-foldr (lambda (x y)
                    (cons x (force y)))
                  '()
                  stream)))

;; filtrace prvku streamu podle vlastnosti
(define stream-filter
  (lambda (prop? stream)
    (stream-foldr (lambda (x y)
                    (if (prop? x)
                        (cons-stream x (force y))
                        (force y)))
                  '()
                  stream)))

;; NEKONECNE A IMPLICITNI PROUDY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;; (potencialne) nekonecny proud jednicek

;; rekurzivni procedura ,,bez limitni podminky''
(define ones-proc
  (lambda ()
    (cons-stream 1 (ones-proc))))

;; nekonecny proud vytvoreny volanim ones-proc
(define ones (ones-proc))

;; predchozi s pouzitim pojmenovaneho ``let''
(define ones (let proc ()
               (cons-stream 1 (proc))))

;; implicitni definice proudu
(define ones (cons-stream 1 ones))

;;;;; (potencialne) nekonecny proud prirozenych cisel

;; rekurzivni procedura ,,bez limitni podminky''
(define naturals-proc
  (lambda (n)
    (cons-stream n (naturals-proc (+ n 1)))))

;; nekonecny proud vytvoreny volanim ones-proc
(define naturals (natural-proc 1))

;; predchozi s pouzitim pojmenovaneho ``let''
(define naturals (let iter ((i 1))
                  (cons-stream i
                               (iter (+ i 1)))))

;; implicitni definice proudu
(define naturals (cons-stream 1
		  (stream-map + ones natural)))

;;;;; proudy sudych a lichych prirozenych cisel

;; lze udelat z proudu prirozenych cisel filtraci
(display-stream (stream-filter even? naturals) 20)
(display-stream (stream-filter odd? naturals) 20)

;; stream hodnot 2^n
(define pow2
  (let next ((last 1))
    (cons-stream last
		 (next (* 2 last )))))

;; implicitni definice predchoziho
(define pow2
  (cons-stream 1
	       (stream-map (lambda (x) (* 2 x))
			   pow2)))
