
;;  Paradigmata Programovani 2B, PRIKLADY
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CYKLY VYBAVENE BREAK, CONTINUE A REDO
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; cyklus typu ``WHILE''
;;;;;;;;;;;;;;;;;;;;;;;;

;; while vybaveny `break'
(define-macro while
  (lambda (condition . body)
    (let ((loop-name (gensym)))
      `(call/cc
	(lambda (break)
	  (let ,loop-name ()
	       (if ,condition
		   (begin ,@body
			  (,loop-name)))))))))

; priklad pouziti:
(let ((i 0)
      (j 0))
  (while (< i 10)
	 (set! j (+ j i))
	 (set! i (+ i 1))
	 (if (> i 5)
	     (break)))
  (list i j))

; priklad vnoreneho pouziti:

;; while vybaveny `break' a `continue'
(define-macro while
  (lambda (condition . body)
    (let ((loop-name (gensym)))
      `(call/cc
	(lambda (break)
	  (let ,loop-name ()
	       (if ,condition
		   (begin (call/cc
			   (lambda (continue)
			     ,@body))
			  (,loop-name)))))))))

; priklad pouziti:
(let ((i 0)
      (j 0))
  (while (< i 10)
	 (if (<= i 2)
	     (begin
	       (set! i (+ i 1))
	       (continue)))
	 (set! j (+ j i))
	 (set! i (+ i 1))
	 (if (> i 5)
	     (break)))
  (list i j))

;; while vybaveny `break', `continue' a `redo' (z jazyka PERL)
;; redo zacne vykonavat telo cyklu od zacatku, ale nevyhodnocuje podminku
(define-macro while
  (lambda (condition . body)
    (let ((loop-name (gensym)))
      `(call/cc
	(lambda (break)
	  (let ,loop-name ()
	       (if ,condition
		   (begin (call/cc
			   (lambda (continue)
			     (let ((redo #f))
			       (call/cc
				(lambda (f)
				  (set! redo f)))
			       ,@body)))
			  (,loop-name)))))))))

;; priklad pouziti:
(let ((i 0)
      (j 0))
  (while (< i 10)
	 (set! j (+ j i))
	 (set! i (+ i 1))
	 (if (and (>= i 10) (< i 20))
	     (redo)))
  (list i j))	; (20 190)

;; pro continue misto `redo' bychom dostali:
(let ((i 0)
      (j 0))
  (while (< i 10)
	 (set! j (+ j i))
	 (set! i (+ i 1))
	 (if (and (>= i 10) (< i 20))
	     (continue)))
  (list i j))	; (10 45)

;; cyklus typu ``DO'' (Scheme) pomoci ``letrec''
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; break, continue a redo pro cyklus typu `do'
(define-macro do
  (lambda (binding condition . body)
    (let ((loop-name (gensym)))
      `(call/cc
	(lambda (break)
	  (letrec ((,loop-name
		    (lambda ,(map car binding)
		      (if ,(car condition)
			  (begin ,@(cdr condition))
			  (begin
			    (call/cc
			     (lambda (continue)
			       (let ((redo #f))
				 (call/cc
				  (lambda (f)
				    (set! redo f)))
				 ,@body)))
			    (,loop-name ,@(map caddr binding)))))))
	    (,loop-name ,@(map cadr binding))))))))

; priklad pouziti (klasika)
(do ((x '(1 3 5 7 9) (cdr x))
     (sum 0 (+ sum (car x))))
    ((null? x) sum)
  (display sum)
  (newline))

; priklad pouziti (break)
(do ((x '(1 3 5 7 9) (cdr x))
     (sum 0 (+ sum (car x))))
    ((null? x) sum)
  (display sum)
  (break)
  (newline))

; priklad pouziti (continue)
(do ((x '(1 3 5 7 9) (cdr x))
     (sum 0 (+ sum (car x))))
    ((null? x) sum)
  (display sum)
  (newline)
  (if (>= sum 10)
      (begin
	(set! sum (- sum))
	(continue))))

; priklad pouziti (redo)
(do ((x '(1 3 5 7 9) (cdr x))
     (sum 0 (+ sum (car x))))
    ((null? x) sum)
  (display sum)
  (newline)
  (if (>= sum 10)
      (begin
	(set! sum (- sum))
	(redo))))

;; cyklus typu ``REPEAT-UNTIL''
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; makro realizujici cyklus typu `REPEAT-UNTIL'
(define-macro repeat
  (lambda args
    (letrec ((but-last
	      (lambda (l)
		(cond ((null? l) #f)
		      ((null? (cdr l)) (cons '() (car l)))
		      (else (let ((result (but-last (cdr l))))
			      (cons (cons (car l) (car result))
				    (cdr result)))))))
	     (split-args (but-last args))
	     (body (car split-args))
	     (limits (cdr split-args))
	     (loop-name (gensym)))
      `(let ,loop-name ()
	    ,@body
	    (cond ,@(map (lambda (conds)
			   `(,(car conds) (begin ,@(cdr conds))))
			 (cdr limits))
		  (else (,loop-name)))))))

;; makro obohacene o `break', `continue' a `redo'
(define-macro repeat
  (lambda args
    (letrec ((but-last
	      (lambda (l)
		(cond ((null? l) #f)
		      ((null? (cdr l)) (cons '() (car l)))
		      (else (let ((result (but-last (cdr l))))
			      (cons (cons (car l) (car result))
				    (cdr result)))))))
	     (split-args (but-last args))
	     (body (car split-args))
	     (limits (cdr split-args))
	     (loop-name (gensym)))
      `(call/cc
	(lambda (break)
	  (let ,loop-name ()
	       (call/cc
		(lambda (continue)
		  (let ((redo #f))
		    (call/cc (lambda (f) (set! redo f)))
		    ,@body)))
	       (cond ,@(map (lambda (conds)
			      `(,(car conds) (begin ,@(cdr conds))))
			    (cdr limits))
		     (else (,loop-name)))))))))

; priklad pouziti (srovnej redo vs. continue)
(let ((x 20)
      (y 15))
  (repeat
   (set! x (- x 1))
   (if (even? x)
       (redo))
   (set! y (+ y 4))
   (until ((<= x 10) (list 'foo x y))
	  ((even? x) 'sude)
	  ((>= y 30) (list 'bar y (+ x 20))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; KORUTINY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; makro na vytvareni korutin
(define-macro coroutine
  (lambda (arg . body)
    `(letrec ((state (lambda ,arg ,@body))
	      (resume
	       (lambda (c . elems)
		 (call/cc
		  (lambda (f)
		    (set! state f)
		    (apply c elems))))))
       (lambda elems
	 (apply state elems)))))

; priklad: (bez pouziti resume se chova jako ,,normalni procedura'')
(define c (coroutine (x) (+ x 1)))
(c 10)	; 11

; priklad: dve korutiny, jedna prepne na druhou
(define d (coroutine ()
	    (display "VOLANA D")
	    (newline)
	    (resume c 10)))

(d)	; 11 + VYPISE `VOLANA d'
(d)	; ... nic (na d je v tuto chvili navazano aktialni pokracovani
        ; vyhodnoceni posledniho vyrazu tela .. to jest pouze zobrazeni
        ; predane hodnoty (PRINT) a opakuje se REPL
(d 10)	  ; 10
(d 'blah) ; blah

; priklad: dve korutiny, ktere se vzajemne prepinaji
(define c (coroutine ()
	    (let iter ((i 3))
	      (display (list "C" i))
	      (newline)
	      (if (> i 0)
		  (begin
		    (display "C")
		    (newline)
		    (resume d)
		    (iter (- i 1)))
		  (begin
		    (display (list "KONEC C"))
		    (newline)
		    'konec-c)))))

(define d (coroutine ()
	    (let iter ((i 3))
	      (display (list "D" i))
	      (newline)
	      (if (> i 0)
		  (begin
		    (display "D")
		    (newline)
		    (resume c)
		    (iter (- i 1)))
		  (begin
		    (display (list "KONEC D"))
		    (newline)
		    'konec-d)))))

(d)	; (D 3)D(C 3)C(D 2)D(C 2)C(D 1)D(C 1)C(D 0) + konec-d je vysledek
(d)	; (D 0)konec-d
(d)	; (D 0)konec-d
(c)	; (C 0)(D 2)D(C 0)(D 2)D ad infinitum:
        ; volani (c) vratilo vypocet do faze korutiny C se stavem i=1,
        ; po dokonceni resume korutiny D, takze probehne jednou cyklus,
        ; vypise se KONEC C + C konci cinnost ... to jest vypocet pokracuje
        ; aktualnim pokracovain konce cinnosti C, coz je bod vypoctu, kde
        ; byla korutina poprve volana, to je D s hodnotou i=3, dojde
        ; k dekrementaci i na 2 + v dalsim cyklu k resumovani C + vse
        ; se opakuje: CYKLIME diky tomu, ze se opakovane vracime k
        ; teze historii vypoctu (aneb: historie se opakuje)

;; nasledujici priklad odstranuje predchozi problem:
(define c (coroutine ()
	    (let iter ((i 3))
	      (display (list "C" i))
	      (newline)
	      (if (> i 0)
		  (begin
		    (display "C")
		    (newline)
		    (resume d)
		    (iter (- i 1)))
		  (begin
		    (display (list "KONEC C"))
		    (newline)
		    (call/cc
		     (lambda (f)
		       (set! state f)
		       (set! resume f)))
		    'konec-c)))))

(define d (coroutine ()
	    (let iter ((i 3))
	      (display (list "D" i))
	      (newline)
	      (if (> i 0)
		  (begin
		    (display "D")
		    (newline)
		    (resume c)
		    (iter (- i 1)))
		  (begin
		    (display (list "KONEC D"))
		    (newline)
		    (call/cc
		     (lambda (f)
		       (set! state f)
		       (set! resume f)))
		    'konec-d)))))

;; predchozi reseni: sice uz necykli, ale (c) nam vraci hodnotu konec-d
;; nastava tedy problem s ,,predavanim vyslednych hodnot''
;; korutiny museji ,,mohou predavat vysledek vypoctu treba pomoci ved. ef.''

;; priklad: tri korutiny, jedna ridi cinnost dalsich dvou:
(define e (coroutine ()
	    (let ((c (coroutine ()
	               (let iter ((i 3))
			 (if (> i 0)
			     (begin
			       (resume e #f)
			       (iter (- i 1)))
			     (resume e 'konec-c)))))
		  (d (coroutine ()
  		       (let iter ((i 3))
			 (if (> i 0)
			     (begin
			       (resume e #f)
			       (iter (- i 1)))
			     (resume e 'konec-d))))))
	      (let iter ()
		(let ((cval (resume c))
		      (dval (resume d)))
		  (if cval
		      (list cval dval)
		      (iter)))))))

(e)	; (konec-c konec-d)

;; PRIKLAD: iteratory pomoci korutin
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; identifikator ukonceni iterace (JIZ MAME)
(define *end-of-iteration* (lambda () #f))

;; predikat indikujici konec iterace (JIZ MAME)
(define finished?
  (lambda (elem)
    (eq? elem *end-of-iteration*)))

;; generuje iterator, jimz bude korutina
;; argument teto korutiny je dalsi korutine: tzv. caller
;; caller je korutina vyuzivajici iterator
(define generate-iterator
  (lambda (l)
    (coroutine (caller)
      (let loop ((l l))
	(if (null? l)
	    (resume caller *end-of-iteration*)
	    (begin
	      (resume caller (car l))
	      (loop (cdr l))))))))

;; k tomu abychom mohli pouzivat iterator potrebujeme vytvorit
;; dalsi korutinu (zde se jmenuje ``user'', v predchozim ``caller'')
(letrec ((iterator (generate-iterator '(a b c d e)))
	 (user
	  (coroutine ()
	    (let iter ()
	      (let ((v (resume iterator user)))
		(if (not (finished? v))
		    (begin
		      (display v)
		      (newline)
		      (iter))))))))
  (user))

;; PRIKLAD: reseni problemu equal-fringe?
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; hloubkovy iterator pomoci korutin
(define generate-depth-iterator
  (lambda (l)
    (coroutine (caller)
      (let loop ((l l))
	(cond ((null? l) 'nejaka-hodnota)
	      ((pair? l)
	       (begin
		 (loop (car l))
		 (loop (cdr l))))
	      (else (resume caller l))))
      (resume caller '()))))

;; equal-fringe? pomoci korutin vyuzivajici iterator
(define equal-fringe?
  (lambda (s1 s2)
    (letrec ((c1 (generate-depth-iterator s1))
	     (c2 (generate-depth-iterator s2))
	     (test (coroutine ()
                     (let iter ()
		       (let ((v1 (resume c1 test))
			     (v2 (resume c2 test)))
			 (or (and (null? v1) (null? v2))
			     (and (equal? v1 v2) (iter))))))))
      (test))))

(equal-fringe? '(a b c d) '(a b c d))		; #t
(equal-fringe? '(a (b c) d) '((a b) c d))	; #t
(equal-fringe? '(a (b (c)) () d) '(a b c (d)))	; #t
(equal-fringe? '(a (b (c)) () d) '(a b c (e)))	; #f
(equal-fringe? '(x (b (c)) () d) '(a b c (d)))	; #f

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; NEDETERMINISMUS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; amb-fail:
;; bez parametru: slouzi k vyvolani navratu (backtracking)
;; s jednim parametrem: nastavi hodnotu aktualniho navratu na dany parametr
(define amb-fail
  (let ((failure #f))
    (lambda args
      (if (null? args)
	  failure
	  (set! failure (if (procedure? (car args))
			    (car args)
			    (lambda ()
			      (error "AMB: Tree Exhausted"))))))))

;; inicializace amb-fail
(amb-fail #f)

; Poznamka: amb-fail je vlastne ,,kontainer''
(amb-fail)	; procedura
((amb-fail))	; Tree Exhausted
(amb-fail 3)
(amb-fail)	; 3

;; amb (ambiguous) operator
(define-macro amb
  (lambda elems
    (let ((previous-fail (gensym)))
      `(let ((,previous-fail (amb-fail)))
	 (call/cc
	  (lambda (exit)
	    ,@(map (lambda (elem)
		     `(call/cc
		       (lambda (next)
			 (amb-fail
			  (lambda ()
			    (amb-fail ,previous-fail)
			    (next)))
			 (exit ,elem))))
		   elems)
	    (,previous-fail)))))))

;; Priklady:
;;;;;;;;;;;;

(amb)

(amb 1 2 3 4)

; zde je videt konvergence, amb nemuze
; vratit #f, protoze mu jde o to, aby nejake reseni nasel
(if (amb #f #t)
    'blah
    (amb))

; zde je taky videt konvergence
(let ((x (amb 1 2 3 4)))
  (if (odd? x)
      (amb)
      x))

;; odvozeny konstrukt: assert
(define assert
  (lambda (elem)
    (if (not elem)
	(amb))))

; priklad pouziti assert (spolu s amb)
(let ((x (amb 1 2 3 4 5)))
  (assert (odd? x))
  x)

;; PROBLEM: BARVENI GRAFU
;;;;;;;;;;;;;;;;;;;;;;;;;

;; opacko ze cviceni: grafy obohatime o informaci o barve v uzlech
(define create-node (lambda (id) (list id 'no-color)))
(define get-neighbors cddr)
(define set-color! (lambda (node color) (set-car! (cdr node) color)))
(define get-color cadr)

(define add-neighbor!
  (lambda (node neighbor)
    (if (not (memq node (get-neighbors neighbor)))
	(begin
	  (set-cdr! (cdr node) (cons neighbor (get-neighbors node)))
	  (set-cdr! (cdr neighbor) (cons node (get-neighbors neighbor)))))))

(define borders '((cz (sk pl de au))
		  (sk (cz pl au hu uk))
		  (pl (de cz sk uk))
		  (de (cz pl au))
		  (hu (au sk uk))
		  (au (cz sk de))
		  (uk (sk pl hu))))

(for-each
 (lambda (x)
   (set-car! x (create-node (car x))))
 borders)

(define countries (map car borders))
countries

(for-each
 (lambda (c)
   (for-each
    (lambda (n)
      (add-neighbor! (car c) (assoc n countries)))
    (cadr c)))
 borders)

countries

; vlastni barveni (trema barvama by neslo):
(amb-fail #f)
(for-each
 (lambda (country)
   (set-color! country (amb 'red 'green 'blue))
   (for-each
    (lambda (neighbor)
      (assert (not (equal? (get-color neighbor)
			   (get-color country)))))
    (get-neighbors country)))
 countries)

countries
 
;; amb by mohl pouzivat imperativni shuffle!, aby nemel pouze jednu
;; strategii prohledavani stavoveho prostoru

;; imperativne zamichej seznam
(define shuffle!
  (lambda (list)
    (let iter ((tail list)
	       (len (length list)))
      (if (not (null? tail))
	  (let ((choose (list-tail tail (random len)))
		(first (car tail)))
	    (set-car! tail (car choose))
	    (set-car! choose first)
	    (iter (cdr tail) (- len 1)))))
    list))

;; amb (ambiguous) operator
(define-macro amb
  (lambda elems
    (let ((previous-fail (gensym))
	  (shuffle!
	   (lambda (list)
	     (let iter ((tail list)
			(len (length list)))
	       (if (not (null? tail))
		   (let ((choose (list-tail tail (random len)))
			 (first (car tail)))
		     (set-car! tail (car choose))
		     (set-car! choose first)
		     (iter (cdr tail) (- len 1)))))
	     list)))
      `(let ((,previous-fail (amb-fail)))
	 (call/cc
	  (lambda (exit)
	    ,@(map (lambda (elem)
		     `(call/cc
		       (lambda (next)
			 (amb-fail
			  (lambda ()
			    (amb-fail ,previous-fail)
			    (next)))
			 (exit ,elem))))
		   (shuffle! elems))
	    (,previous-fail)))))))
